// ============================================================================
#ifndef CALOALGS_CALOSINGLEPHOTONALG_H 
#define CALOALGS_CALOSINGLEPHOTONALG_H 1
// ============================================================================
#include <string>
#include "GaudiAlg/ScalarTransformer.h"
#include "CaloDet/DeCalorimeter.h"
#include "Event/CaloHypo.h"
#include "Event/CaloCluster.h"
#include "CaloInterfaces/ICounterLevel.h"
struct ICaloClusterSelector ;
struct ICaloHypoTool        ;

/** @class CaloSinglePhotonAlg CaloSinglePhotonAlg.h
 *  
 *  The simplest algorithm of reconstruction of 
 *  single photon in electromagnetic calorimeter.
 *  The implementation is based on F.Machefert's codes.
 *
 *  @author Frederic Machefert machefer@in2p3.fr
 *  @author Vanya Belyaev      Ivan.Belyaev@itep.ru
 *  @date   31/03/2002
 */
class CaloSinglePhotonAlg 
: public Gaudi::Functional::ScalarTransformer<CaloSinglePhotonAlg,LHCb::CaloHypos(const LHCb::CaloClusters&)>
{  
  
public:
  CaloSinglePhotonAlg( const std::string&  name , ISvcLocator* pSvc );  

  StatusCode initialize () override;  
  StatusCode finalize   () override;

  using ScalarTransformer::operator();
  boost::optional<LHCb::CaloHypo> operator()(const LHCb::CaloCluster& ) const;
  void postprocess(const LHCb::CaloHypos&) const;
  
private:
  
  /// container of names
  typedef std::vector<std::string>           Names       ;
  /// container of selector tools 
  typedef std::vector<ICaloClusterSelector*> Selectors   ;
  /// containers of hypo tools 
  typedef std::vector<ICaloHypoTool*>        HypoTools   ;
  /// container of correction tools (S-,L-,...)
  typedef HypoTools                          Corrections ;
    
  // cluster selectors 
  Selectors m_selectors;
  Gaudi::Property<Names> m_selectorsTypeNames 
    {this, "SelectionTools", { 
      "CaloSelectCluster/PhotonCluster",
      "CaloSelectNeutralClusterWithTracks/NeutralCluster"
    }, "List of tools for selection of clusters"};

  // corrections
  Corrections  m_corrections;
  Gaudi::Property<Names> m_correctionsTypeNames 
    {this, "CorrectionTools", {}, 
    "List of tools for primary corrections"};
  
  // other hypo tools 
  HypoTools m_hypotools;
  Gaudi::Property<Names> m_hypotoolsTypeNames 
    {this, "HypoTools", 
    { "CaloExtraDigits/SpdPrsExtraG" },
    "List of generic Hypo-tools to apply to newly created hypos"};

  // corrections
  Corrections m_corrections2;
  Gaudi::Property<Names> m_correctionsTypeNames2 
    {this, "CorrectionTools2", { 
      "CaloECorrection/ECorrection",
      "CaloSCorrection/SCorrection",
      "CaloLCorrection/LCorrection", 
    }, "List of tools for 'fine-corrections' "};

  // other hypo tools 
  HypoTools m_hypotools2;
  Gaudi::Property<Names> m_hypotoolsTypeNames2 
    {this, "HypoTools2", {}, "List of generic Hypo-tools to apply to corrected hypos"};

  Gaudi::Property<std::string> m_detData 
    {this, "Detector", LHCb::CaloAlgUtils::DeCaloLocation("Ecal")};
  
  Gaudi::Property<double> m_eTcut 
    {this, "EtCut", 0., "Threshold on cluster & hypo ET"};
  
  LHCb::CaloDataFunctor::EnergyTransverse<const DeCalorimeter*> m_eT { nullptr } ; // TODO: can we combine this into Over_Et_Threshold and make it a property??

  const DeCalorimeter* m_det = nullptr;
  ICounterLevel* counterStat = nullptr;
};
// ============================================================================
#endif // CALOSINGLEPHOTONALG_H
