/** Implementation of MuonIDAlgLite.
 *
 * 2015-06-01: Kevin Dungs, Ricardo Vazquez Gomez
 * Updated to functional:
 * 2017-01-26: Holger Stevens, Martin Bieker, Julian Surmann
 * 2017-??-??: Roel Aaij
 * 2017-12-12: Nikita Kazeev
 */
#include "MuonIDAlgLite.h"

DECLARE_COMPONENT( MuonIDAlgLite )

/** Load the CommonMuonTool.
 */
MuonIDAlgLite::MuonIDAlgLite(const std::string &name,
			     ISvcLocator *pSvcLocator) :
MultiTransformer(name, pSvcLocator,
                 // Inputs
                 {KeyValue{"TracksLocations", LHCb::TrackLocation::Default},
           	  KeyValue{"HitHandler", MuonHitHandlerLocation::Default}},
                 // Outputs
                 {KeyValue{"MuonIDLocation", LHCb::MuonPIDLocation::Default},
                  KeyValue{"MuonTrackLocation", LHCb::TrackLocation::Muon}})
{
}

StatusCode MuonIDAlgLite::initialize() {
  auto sc = MultiTransformer::initialize();
  if ( sc.isFailure() ) return sc;
  if ( msgLevel(MSG::DEBUG) ) debug() << "==> Initialize" << endmsg;

  muonTool_ = tool<ICommonMuonTool>("CommonMuonTool");
  DLLTool_ = tool<DLLMuonTool>("DLLMuonTool");
  makeMuonTool_ = tool<MakeMuonTool>("MakeMuonTool");
  MVATool_ = tool<IMVATool>("MVATool");
  return StatusCode::SUCCESS;
}

/** Iterates over all tracks in the current event and performs muon id on them.
 * Resulting PID objects as well as muon tracks are stored on the TES.
 */
std::tuple<LHCb::MuonPIDs,LHCb::Tracks> MuonIDAlgLite::operator()(const LHCb::Tracks& tracks,
                                                                  const MuonHitHandler& hitHandler) const {
  LHCb::MuonPIDs muPids;
  LHCb::Tracks muTracks;
  std::map<LHCb::MuonPID*, float> m_muonDistMap;
  std::map<LHCb::MuonPID*, CommonConstMuonHits> m_muonMap;

  muPids.setVersion(1);  // TODO: still valid? do we need a new version?

  counter("nMomentumCut");
  counter("nPreSelTrack");
  counter("nGoodOffline");
  counter("nMuonPIDs");
  counter("nInAcceptance");
  counter("nIsMuon");
  counter("nIsMuonLoose");
  counter("nIsMuonTight");
  // Iterate over all tracks and do the offline identification for each of them
  for(const auto newTrackPtr: tracks) {
    const auto &track = *newTrackPtr;
    if (!isGoodOfflineTrack(track)) {
      continue;
    }
    counter("nGoodOffline")++;
    auto muPid = std::make_unique<LHCb::MuonPID>(); // TODO: Put to where the values are set after the
                                                    // muonMap thing has been figured out...
    muPid->setIDTrack(newTrackPtr);
    muPid->setIsMuon(false);
    muPid->setIsMuonLoose(false);
    muPid->setIsMuonTight(false);
    muPid->setNShared(0);
    muPid->setPreSelMomentum(false);
    muPid->setInAcceptance(true);
    muPid->setMuonLLMu(-10000.);
    muPid->setMuonLLBg(-10000.);
    muPid->setChi2Corr(0);
    muPid->setMuonMVA1(-999.);
    muPid->setMuonMVA2(-999.);
    muPid->setMuonMVA3(-999.);
    muPid->setMuonMVA4(-999.);

    double probMu = 0.00000000001, probNonMu = 0.00000000001;
    double Dsquare = -1;

    bool preSel = muonTool_->preSelection(track);
    const auto extrapolation = muonTool_->extrapolateTrack(track);
    if (preSel) {
      muPid->setPreSelMomentum(1);
      ++counter("nMomentumCut");
    }
    // Store the muonPIDs only for tracks passing InAcceptance
    if (!muonTool_->inAcceptance(extrapolation)) {
      continue;
    }
    muPid->setInAcceptance(1);
    ++counter("nInAcceptance");


    CommonConstMuonHits hits, hitsTight;
    if (preSel) {
      counter("nPreSelTrack")++;

      ICommonMuonTool::MuonTrackOccupancies occupancies, occupanciesTight;
      std::tie(hits, occupancies) =
         muonTool_->hitsAndOccupancies(track, extrapolation, hitHandler);
      std::tie(hitsTight, occupanciesTight) = muonTool_->extractCrossed(hits);
      const auto isMuon = muonTool_->isMuon(occupancies, track.p());
      const auto isMuonTight = muonTool_->isMuon(occupanciesTight, track.p());
      const auto isMuonLoose = muonTool_->isMuonLoose(occupancies, track.p());
      muPid->setIsMuon(isMuon);
      muPid->setIsMuonLoose(isMuonLoose);
      muPid->setIsMuonTight(isMuonTight);
      if (isMuon) {
        counter("nIsMuon")++;
      }
      if (isMuonTight) {
        counter("nIsMuonTight")++;
      }

      if (isMuonLoose and runDLL_) {
        counter("nIsMuonLoose")++;

        // TODO: This has to go
        if (m_dllFlag == 4) {
          std::tie(probMu, probNonMu, Dsquare) = DLLTool_->calcMuonLL_tanhdist_landau(
              track, extrapolation, hits, occupancies);
        } else if (m_dllFlag == 5) {
          std::tie(probMu, probNonMu, Dsquare) = DLLTool_->calcMuonLL_tanhdist(
              track, extrapolation, hits, occupancies);
        } else {
          Error("Invalid DLL flag").ignore();  // TODO: seriously this is just a quick fix,
          // this whole flag thing has to go
        }

        muPid->setMuonLLMu(log(probMu));
        muPid->setMuonLLBg(log(probNonMu));

        // TODO(kazeevn) check return value
        DLLTool_->calcNShared(
			      muPid.get(), &muPids, hits, extrapolation, &m_muonDistMap,
                              &m_muonMap);
      }
    }
    muPids.insert(muPid.get());  //ownership goes here...
    counter("nMuonPIDs")++;

    // make the muon track only for tracks passing IsMuonLoose
    if (muPid->IsMuonLoose()) {
      auto muTrack = makeMuonTool_->makeMuonTrack(muPid.get(), hits, extrapolation);

      // Insert in muonTrack
      muTrack->setType(LHCb::Track::Types::Muon);
      muTrack->setHistory(LHCb::Track::History::MuonID);
      muTrack->addInfo(LHCb::Track::AdditionalInfo::MuonMomentumPreSel, muPid->PreSelMomentum());
      muTrack->addInfo(LHCb::Track::AdditionalInfo::MuonInAcceptance, muPid->InAcceptance());
      muTrack->addInfo(LHCb::Track::AdditionalInfo::IsMuonLoose, muPid->IsMuonLoose());
      muTrack->addInfo(LHCb::Track::AdditionalInfo::IsMuon, muPid->IsMuon());
      muTrack->addInfo(LHCb::Track::AdditionalInfo::IsMuonTight, muPid->IsMuonTight());
      muTrack->addInfo(LHCb::Track::AdditionalInfo::MuonDist2, Dsquare);
      muTrack->addInfo(LHCb::Track::AdditionalInfo::MuonDLL, muPid->MuonLLMu() - muPid->MuonLLBg());
      muTrack->addInfo(LHCb::Track::AdditionalInfo::MuonNShared, muPid->nShared());
      //muTrack->addInfo(LHCb::Track::AdditionalInfo::MuonChi2perDoF, muTrack->chi2PerDoF());
      muPid->setMuonTrack(muTrack);
      muTracks.insert(muTrack);

      if (runMVA_ && muPid->IsMuon()) {
        // compute the MVA variable
        // Only compute the MVA for tracks passing isMuon, as the BDT has been trained with this cut.
        // This follows the same logic as the ProbNN which is not computed
     	  // for tracks with IsMuonLoose == 1 and IsMuon == 0
	      IMVATool::MVAResponses mva = MVATool_->calcBDT(*muTrack, hits);
	      if (MVATool_->getRunTMVA()) {
	        muPid->setMuonMVA1(mva.TMVA);
	      }
	      if (MVATool_->getRunCatBoost()) {
	        muPid->setMuonMVA2(mva.CatBoost);
	      }
      }
    }

    muPid.release(); // ownership goes with muPids

  }
  return std::make_tuple(std::move(muPids), std::move(muTracks));
}


/** Offline requirement for a good track. The track mustn't be a clone. Also it
 * should either be a long track or a downstream track or in case TT tracks are
 * supposed to be used a TT track.
 */
bool MuonIDAlgLite::isGoodOfflineTrack(const LHCb::Track &track) const {
  return !track.checkFlag(LHCb::Track::Flags::Clone) &&
         (track.checkType(LHCb::Track::Types::Long) ||
          track.checkType(LHCb::Track::Types::Downstream) ||
          (useTTrack_ && track.checkType(LHCb::Track::Types::Ttrack)));
}
