
// local
#include "RichMCSIMDPhotonCherenkovAngles.h"

using namespace Rich::Future::Rec::MC::Moni;

//-----------------------------------------------------------------------------
// Implementation file for class : SIMDPhotonCherenkovAngles
//
// 2016-12-06 : Chris Jones
//-----------------------------------------------------------------------------

SIMDPhotonCherenkovAngles::
SIMDPhotonCherenkovAngles( const std::string& name, ISvcLocator* pSvcLocator )
  : Consumer( name, pSvcLocator,
              { KeyValue{ "SummaryTracksLocation", Summary::TESLocations::Tracks },
                KeyValue{ "TracksLocation",          LHCb::TrackLocation::Default },
                KeyValue{ "RichSIMDPixelSummariesLocation", SIMDPixelSummariesLocation::Default },
                KeyValue{ "RichPixelClustersLocation", Rich::PDPixelClusterLocation::Default },
                KeyValue{ "PhotonToParentsLocation", Relations::PhotonToParentsLocation::Default },
                KeyValue{ "TrackSegmentsLocation", LHCb::RichTrackSegmentLocation::Default },
                KeyValue{ "CherenkovAnglesLocation", CherenkovAnglesLocation::Signal },
                KeyValue{ "CherenkovPhotonLocation", SIMDCherenkovPhotonLocation::Default },
                KeyValue{ "TrackToMCParticlesRelations", Rich::Future::MC::Relations::TrackToMCParticles },
                KeyValue{ "RichDigitSummariesLocation", LHCb::MCRichDigitSummaryLocation::Default } } )
{
  // print some stats on the final plots
  setProperty ( "HistoPrint", true ); 
}

//-----------------------------------------------------------------------------

StatusCode SIMDPhotonCherenkovAngles::prebookHistograms()
{

  // Loop over radiators
  for ( const auto rad : Rich::radiators() )
  {
    if ( m_rads[rad] )
    {

      // CkRes
      richHisto1D( HID("ckResAll",rad), "Rec-Exp Cktheta | True Type",
                   -m_ckResRange[rad], m_ckResRange[rad], nBins1D() );
      richHisto1D( HID("ckResTrue",rad), "Rec-Exp Cktheta | MC true photons | True Type",
                   -m_ckResRange[rad], m_ckResRange[rad], nBins1D() );
      richHisto1D( HID("thetaRecTrue",rad), "Reconstructed Ch Theta | MC true photons",
                   m_ckThetaMin[rad], m_ckThetaMax[rad], nBins1D(),
                   "Cherenkov Theta / rad" );
      richHisto1D( HID("phiRecTrue",rad), "Reconstructed Ch Phi | MC true photons",
                   0.0, 2.0*Gaudi::Units::pi, nBins1D(),
                   "Cherenkov Phi / rad" );
      richHisto1D( HID("thetaRecFake",rad), "Reconstructed Ch Theta | MC fake photons",
                   m_ckThetaMin[rad], m_ckThetaMax[rad], nBins1D(),
                   "Cherenkov Theta / rad" );
      richHisto1D( HID("phiRecFake",rad), "Reconstructed Ch Phi | MC fake photons",
                   0.0, 2.0*Gaudi::Units::pi, nBins1D(),
                   "Cherenkov Phi / rad" );
      richHisto1D( HID("ckResFake",rad), "Rec-Exp Cktheta | MC fake photons | True Type",
                   -m_ckResRange[rad], m_ckResRange[rad], nBins1D() );
      richProfile1D( HID("ckResTrueVP",rad), "<Rec-Exp Cktheta> V P | MC true photons | True Type",
                     m_minP, m_maxP, nBins1D(),
                     "Track Momentum (MeV/c)", "<Rec-Exp Cktheta> / rad" );

      // Plots assuming pion hypo
      richHisto1D( HID("ckResAllPion",rad), "Rec-Exp Cktheta | Pion Type", 
                   -m_ckResRange[rad], m_ckResRange[rad], nBins1D() );
      richHisto1D( HID("ckResTruePion",rad), "Rec-Exp Cktheta | MC true photons | Pion Type",
                   -m_ckResRange[rad], m_ckResRange[rad], nBins1D() );
      richHisto1D( HID("ckResFakePion",rad), "Rec-Exp Cktheta | MC fake photons | Pion Type",
                   -m_ckResRange[rad], m_ckResRange[rad], nBins1D() );

      // Photon hit time
      richHisto1D( HID("timeAll",rad), "Photon time | True Type",
                   m_timeLow[rad], m_timeHigh[rad], nBins1D() );
      richHisto1D( HID("timeTrue",rad), "Photon time | MC true photons | True Type",
                   m_timeLow[rad], m_timeHigh[rad], nBins1D() );
      richHisto1D( HID("timeFake",rad), "Photon time | MC fake photons | True Type",
                   m_timeLow[rad], m_timeHigh[rad], nBins1D() );

      // Photon hit time (wider range)
      richHisto1D( HID("timeAll-wide",rad), "Photon time | True Type",
                   m_timeLowWide[rad], m_timeHighWide[rad], nBins1D() );
      richHisto1D( HID("timeTrue-wide",rad), "Photon time | MC true photons | True Type",
                   m_timeLowWide[rad], m_timeHighWide[rad], nBins1D() );
      richHisto1D( HID("timeFake-wide",rad), "Photon time | MC fake photons | True Type",
                   m_timeLowWide[rad], m_timeHighWide[rad], nBins1D() );

      // Photon hit time and trackPath IF difference is not on the main peak...
      richHisto1D( HID("secondPeakHitTimeAll",rad), "Photon time if time difference 2.5<t<6ns | True Type",
                   m_timeLowWide[rad], m_timeHighWide[rad], nBins1D() );
      richHisto1D( HID("secondPeakHitTimeTrue",rad), "Photon time if time difference 2.5<t<6ns | MC true photons | True Type",
                   m_timeLowWide[rad], m_timeHighWide[rad], nBins1D() );
      richHisto1D( HID("secondPeakHitTimeFake",rad), "Photon time if time difference 2.5<t<6ns | MC fake photons | True Type",
                   m_timeLowWide[rad], m_timeHighWide[rad], nBins1D() );

      // Track origin time
      richHisto1D( HID("trackOriginTimeTrue",rad), "Track Origin time | MC true photons | True Type",
                   -m_timeOrigin[rad], m_timeOrigin[rad], nBins1D() );

      // PV origin time
      richHisto1D( HID("pvOriginTimeTrue",rad), "PV Origin time | MC true photons | True Type",
                   -m_timeOrigin[rad], m_timeOrigin[rad], nBins1D() );

      // Time differences hit-expected
      richHisto1D( HID("diffHitExpTimeAll",rad), "Rich hit t - track t - path length/c | True Type",
                   -m_timeOriginWide[rad], m_timeOriginWide[rad], nBins1D() );
      richHisto1D( HID("diffHitExpTimeTrue",rad), "Rich hit t - track t - path length/c | MC true photons | True Type",
                   -m_timeOriginWide[rad], m_timeOriginWide[rad], nBins1D() );
      richHisto1D( HID("diffHitExpTimeFake",rad), "Rich hit t - track t - path length/c | MC fake photons | True Type",
                   -m_timeOriginWide[rad], m_timeOriginWide[rad], nBins1D() );

      // Time differences track-PV
      richHisto1D( HID("diffTrackPVTimeTrue",rad), "Track-PV Origin time | MC true photons | True Type",
                   -m_timeOrigin[rad], m_timeOrigin[rad], nBins1D() );

      // Track time plus path length
      richHisto1D( HID("trackPathAll",rad), "Track time + path length/c | True Type",
                   -m_timeLowWide[rad], m_timeHighWide[rad], nBins1D() );
      richHisto1D( HID("trackPathTrue",rad), "Track time + path length/c | MC true photons | True Type",
                   -m_timeLowWide[rad], m_timeHighWide[rad], nBins1D() );
      richHisto1D( HID("trackPathFake",rad), "Track time + path length/c | MC fake photons | True Type",
                   -m_timeLowWide[rad], m_timeHighWide[rad], nBins1D() );

      // Photon path length
      richHisto1D( HID("photonPathAll",rad), "Photon path length | True Type",
                   m_photonPathLow[rad], m_photonPathHigh[rad], nBins1D() );
      richHisto1D( HID("photonPathTrue",rad), "Photon path length | MC true photons | True Type",
                   m_photonPathLow[rad], m_photonPathHigh[rad], nBins1D() );
      richHisto1D( HID("photonPathFake",rad), "Photon path length | MC fake photons | True Type",
                   m_photonPathLow[rad], m_photonPathHigh[rad], nBins1D() );

      // Photon path length IF difference is not on the main peak...
      richHisto1D( HID("secondPeakPhotonPathAll",rad), "Photon path length if time difference 2.5<t<6ns | True Type",
                   m_photonPathLow[rad], m_photonPathHigh[rad], nBins1D() );
      richHisto1D( HID("secondPeakPhotonPathTrue",rad), "Photon path length if time difference 2.5<t<6ns | MC true photons | True Type",
                   m_photonPathLow[rad], m_photonPathHigh[rad], nBins1D() );
      richHisto1D( HID("secondPeakPhotonPathFake",rad), "Photon path length if time difference 2.5<t<6ns | MC fake photons | True Type",
                   m_photonPathLow[rad], m_photonPathHigh[rad], nBins1D() );

      // Time vs momentum
      richProfile2D( HID("timeMomentumAll",rad), "Time diff against momentum | True Type",
                     m_zeros[rad], m_momentumHigh[rad], nBins1D(),
                     -m_timeOriginWide[rad], m_timeOriginWide[rad], nBins1D() );
      richProfile2D( HID("timeMomentumAllPion",rad), "Time diff against momentum | Pion Type",
                     m_zeros[rad], m_momentumHigh[rad], nBins1D(),
                     -m_timeOriginWide[rad], m_timeOriginWide[rad], nBins1D() );
      richProfile2D( HID("timeMomentumTrue",rad), "Time diff against momentum | MC true photons | True Type",
                     m_zeros[rad], m_momentumHigh[rad], nBins1D(),
                     -m_timeOriginWide[rad], m_timeOriginWide[rad], nBins1D() );
      richProfile2D( HID("timeMomentumTruePion",rad), "Time diff against momentum | MC true photons | Pion Type",
                     m_zeros[rad], m_momentumHigh[rad], nBins1D(),
                     -m_timeOriginWide[rad], m_timeOriginWide[rad], nBins1D() );
      richProfile2D( HID("timeMomentumFake",rad), "Time diff against momentum | MC fake photons | True Type",
                     m_zeros[rad], m_momentumHigh[rad], nBins1D(),
                     -m_timeOriginWide[rad], m_timeOriginWide[rad], nBins1D() );
      richProfile2D( HID("timeMomentumFakePion",rad), "Time diff against momentum | MC fake photons | Pion Type",
                     m_zeros[rad], m_momentumHigh[rad], nBins1D(),
                     -m_timeOriginWide[rad], m_timeOriginWide[rad], nBins1D() );

      // track+path vs momentum
      richProfile2D( HID("trackPathMomentumAll",rad), "TrackTime+path/c against momentum | True Type",
                     m_zeros[rad], m_momentumHigh[rad], nBins1D(),
                     -m_timeOriginWide[rad], m_timeOriginWide[rad], nBins1D() );
      richProfile2D( HID("trackPathMomentumAllPion",rad), "TrackTime+path/c against momentum | Pion Type",
                     m_zeros[rad], m_momentumHigh[rad], nBins1D(),
                     -m_timeOriginWide[rad], m_timeOriginWide[rad], nBins1D() );
      richProfile2D( HID("trackPathMomentumTrue",rad), "TrackTime+path/c against momentum | MC true photons | True Type",
                     m_zeros[rad], m_momentumHigh[rad], nBins1D(),
                     -m_timeOriginWide[rad], m_timeOriginWide[rad], nBins1D() );
      richProfile2D( HID("trackPathMomentumTruePion",rad), "TrackTime+path/c against momentum | MC true photons | Pion Type",
                     m_zeros[rad], m_momentumHigh[rad], nBins1D(),
                     -m_timeOriginWide[rad], m_timeOriginWide[rad], nBins1D() );
      richProfile2D( HID("trackPathMomentumFake",rad), "TrackTime+path/c against momentum | MC fake photons | True Type",
                     m_zeros[rad], m_momentumHigh[rad], nBins1D(),
                     -m_timeOriginWide[rad], m_timeOriginWide[rad], nBins1D() );
      richProfile2D( HID("trackPathMomentumFakePion",rad), "TrackTime+path/c against momentum | MC fake photons | Pion Type",
                     m_zeros[rad], m_momentumHigh[rad], nBins1D(),
                     -m_timeOriginWide[rad], m_timeOriginWide[rad], nBins1D() );

      // Loop over all particle codes
      //for ( const auto hypo : activeParticlesNoBT() ) { }
    }
  }
 
  return StatusCode::SUCCESS;
}

//-----------------------------------------------------------------------------

void 
SIMDPhotonCherenkovAngles::operator()( const Summary::Track::Vector& sumTracks,
                                       const LHCb::Track::Selection& tracks,
                                       const SIMDPixelSummaries& pixels,
                                       const Rich::PDPixelCluster::Vector& clusters,
                                       const Relations::PhotonToParents::Vector& photToSegPix,
                                       const LHCb::RichTrackSegment::Vector& segments,
                                       const CherenkovAngles::Vector& expTkCKThetas,
                                       const SIMDCherenkovPhoton::Vector& photons,
                                       const Rich::Future::MC::Relations::TkToMCPRels& tkrels,
                                       const LHCb::MCRichDigitSummarys & digitSums ) const
{

  // Make a local MC helper object
  Helper mcHelper( tkrels, digitSums );

  // loop over the track info
  for ( const auto && data : Ranges::ConstZip(sumTracks,tracks) )
  {
    const auto & sumTk = std::get<0>(data);
    const auto & tk    = std::get<1>(data);

    // loop over photons for this track
    for ( const auto photIn : sumTk.photonIndices() )
    {
      // photon data
      const auto & phot = photons[photIn];
      const auto & rels = photToSegPix[photIn];

      // Get the SIMD summary pixel
      const auto & simdPix = pixels[rels.pixelIndex()];

      // the segment for this photon
      const auto & seg = segments[rels.segmentIndex()];

      // Radiator info
      const auto rad = seg.radiator();
      if ( !m_rads[rad] ) continue;

      // get the expected CK theta values for this segment
      const auto & expCKangles = expTkCKThetas[rels.segmentIndex()];

      // Segment momentum
      const auto pTot = seg.bestMomentumMag();

      // Track start time
      const auto trackTime = seg.trackOrigin();

      // Track PV time
      const auto pvTime = seg.pvOrigin();

      // Get the MCParticles for this track
      const auto mcPs = mcHelper.mcParticles(*tk,true,0.5);

      // Weight per MCP
      const auto mcPW = ( !mcPs.empty() ? 1.0/(double)mcPs.size() : 1.0 );

      // Loop over scalar entries in SIMD photon
      for ( std::size_t i = 0; i < SIMDCherenkovPhoton::SIMDFP::Size; ++i )
      {
        // Select valid entries
        if ( !phot.validityMask()[i] ) continue;

        // scalar cluster
        const auto & clus = clusters[ simdPix.scClusIndex()[i] ];

	// get smart ID
	const auto & id = phot.smartID()[i];

	// reconstructed path length
	const auto pathLengthRec = phot.pathLength()[i];

	// photon path length 
	const auto photonPathRec = phot.photonPathLength()[i];

        // do we have an true MC Cherenkov photon
        const auto trueCKMCPs = mcHelper.trueCherenkovPhoton( *tk, rad, clus );

        // reconstructed theta
        const auto thetaRec = phot.CherenkovTheta()[i];
       
        // reconstructed phi
        const auto phiRec   = phot.CherenkovPhi()[i];
        
        // loop over MCPs
        for ( const auto mcP : mcPs )
        {
          
          // The True MCParticle type
          const auto pid = mcHelper.mcParticleType(mcP);
          // If MC type not known, abort
          if ( Rich::Unknown == pid ) continue; 
          //auto pid = mcHelper.mcParticleType(mcP);
          // If MC type not known, assume Pion (as in real data)
          //if ( Rich::Unknown == pid ) { pid = Rich::Pion; }

          // skip electrons which are reconstructed badly..
          if ( UNLIKELY( m_skipElectrons && Rich::Electron == pid ) ) { continue; } 

          // beta cut for true MC type
          const auto mcbeta = richPartProps()->beta( pTot, pid );
          if ( mcbeta >= m_minBeta[rad] && mcbeta <= m_maxBeta[rad] )
          {
            
            // true Cherenkov signal ?
            const bool trueCKSig = std::find( trueCKMCPs.begin(), 
                                              trueCKMCPs.end(), mcP ) != trueCKMCPs.end();

            // expected CK theta ( for true type )
            const auto thetaExp = expCKangles[pid];

            // delta theta
            const auto deltaTheta = thetaRec - thetaExp;

	    // Time
	    const auto photonTime = id.ADCNfloat<16>( id.time(),0.,100. );
            
            // fill some plots
            richHisto1D( HID("ckResAll",rad) ) -> fill( deltaTheta, mcPW );
	    richHisto1D( HID("timeAll",rad) ) -> fill( photonTime, mcPW );
	    richHisto1D( HID("timeAll-wide",rad) ) -> fill( photonTime, mcPW );
	    richHisto1D( HID("diffHitExpTimeAll",rad) ) -> fill( (photonTime-trackTime.T()-pathLengthRec/Gaudi::Units::c_light),mcPW );
	    richHisto1D( HID("trackPathAll",rad) ) -> fill( (trackTime.T()+pathLengthRec/Gaudi::Units::c_light),mcPW );
	    richHisto1D( HID("photonPathAll",rad) ) -> fill( photonPathRec,mcPW );
	    richProfile2D( HID("timeMomentumAll",rad) ) -> fill( pTot, (photonTime-trackTime.T()-pathLengthRec/Gaudi::Units::c_light),mcPW );
	    richProfile2D( HID("trackPathMomentumAll",rad) ) -> fill( pTot, (trackTime.T()-pathLengthRec/Gaudi::Units::c_light),mcPW );
            if ( !trueCKSig )
            {
              richHisto1D( HID("ckResFake",rad) ) -> fill( deltaTheta, mcPW );
	      richHisto1D( HID("timeFake",rad) ) -> fill( photonTime, mcPW );
	      richHisto1D( HID("timeFake-wide",rad) ) -> fill( photonTime, mcPW );
	      richHisto1D( HID("diffHitExpTimeFake",rad) ) -> fill( (photonTime-trackTime.T()-pathLengthRec/Gaudi::Units::c_light),mcPW );
	      richHisto1D( HID("trackPathFake",rad) ) -> fill( (trackTime.T()+pathLengthRec/Gaudi::Units::c_light),mcPW );
	      richHisto1D( HID("photonPathFake",rad) ) -> fill( photonPathRec,mcPW );
	      richProfile2D( HID("timeMomentumFake",rad) ) -> fill( pTot, (photonTime-trackTime.T()-pathLengthRec/Gaudi::Units::c_light),mcPW );
	      richProfile2D( HID("trackPathMomentumFake",rad) ) -> fill( pTot, (trackTime.T()-pathLengthRec/Gaudi::Units::c_light),mcPW );
            }
            else
            {

              richHisto1D( HID("trackOriginTimeTrue",rad) ) -> fill( trackTime.T(), mcPW );
              richHisto1D( HID("diffTrackPVTimeTrue",rad) ) -> fill( trackTime.T()-pvTime.T(), mcPW );
	      richHisto1D( HID("diffHitExpTimeTrue",rad) ) -> fill( (photonTime-trackTime.T()-pathLengthRec/Gaudi::Units::c_light),mcPW );
              richHisto1D( HID("pvOriginTimeTrue",rad) ) -> fill( pvTime.T(), mcPW );
	      richHisto1D( HID("timeTrue",rad) ) -> fill( photonTime, mcPW );
	      richHisto1D( HID("timeTrue-wide",rad) ) -> fill( photonTime, mcPW );
	      richHisto1D( HID("trackPathTrue",rad) ) -> fill( (trackTime.T()+pathLengthRec/Gaudi::Units::c_light),mcPW );
	      richHisto1D( HID("photonPathTrue",rad) ) -> fill( photonPathRec,mcPW );
	      richProfile2D( HID("timeMomentumTrue",rad) ) -> fill( pTot, (photonTime-trackTime.T()-pathLengthRec/Gaudi::Units::c_light),mcPW );
	      richProfile2D( HID("trackPathMomentumTrue",rad) ) -> fill( pTot, (trackTime.T()-pathLengthRec/Gaudi::Units::c_light),mcPW );

              richHisto1D  ( HID("ckResTrue",rad)    ) -> fill( deltaTheta, mcPW );
              richHisto1D  ( HID("thetaRecTrue",rad) ) -> fill( thetaRec,   mcPW );
              richHisto1D  ( HID("phiRecTrue",rad)   ) -> fill( phiRec,     mcPW );
              richProfile1D( HID("ckResTrueVP",rad)  ) -> fill( pTot, deltaTheta, mcPW );

            }
	    if( (photonTime-trackTime.T()-pathLengthRec/Gaudi::Units::c_light)>=2.5 && (photonTime-trackTime.T()-pathLengthRec/Gaudi::Units::c_light)<=6.0 ) {
	      richHisto1D( HID("secondPeakHitTimeAll",rad) ) -> fill( photonTime, mcPW );
	      richHisto1D( HID("secondPeakPhotonPathAll",rad) ) -> fill( photonPathRec,mcPW );
	      if ( !trueCKSig ) {
		richHisto1D( HID("secondPeakHitTimeFake",rad) ) -> fill( photonTime, mcPW );
		richHisto1D( HID("secondPeakPhotonPathFake",rad) ) -> fill( photonPathRec,mcPW );
	      } else {
		richHisto1D( HID("secondPeakHitTimeTrue",rad) ) -> fill( photonTime, mcPW );
		richHisto1D( HID("secondPeakPhotonPathTrue",rad) ) -> fill( photonPathRec,mcPW );
	      }
	    }
            
          }
          
        } // loop over associated MCPs
        
        // Now plots when assuming all tracks are pions ( as in real data )
        const auto pionbeta = richPartProps()->beta( pTot, Rich::Pion );
        if ( pionbeta >= m_minBeta[rad] && pionbeta <= m_maxBeta[rad] )
        {
          
          // expected CK theta ( for Pion )
          const auto thetaExp = expCKangles[Rich::Pion];
          
          // delta theta
          const auto deltaTheta = thetaRec - thetaExp;
          
	  // Time
	  const auto photonTime = id.ADCNfloat<16>( id.time(),0.,100. );
	  
          // fill some plots   
          richHisto1D( HID("ckResAllPion",rad) ) -> fill( deltaTheta );
	  richProfile2D( HID("timeMomentumAllPion",rad) ) -> fill( pTot, (photonTime-trackTime.T()-pathLengthRec/Gaudi::Units::c_light),mcPW );
	  richProfile2D( HID("trackPathMomentumAllPion",rad) ) -> fill( pTot, (trackTime.T()-pathLengthRec/Gaudi::Units::c_light),mcPW );
          if ( trueCKMCPs.empty() )
          {
            richHisto1D( HID("ckResFakePion",rad) ) -> fill( deltaTheta );
	    richProfile2D( HID("timeMomentumFakePion",rad) ) -> fill( pTot, (photonTime-trackTime.T()-pathLengthRec/Gaudi::Units::c_light),mcPW );
	    richProfile2D( HID("trackPathMomentumFakePion",rad) ) -> fill( pTot, (trackTime.T()-pathLengthRec/Gaudi::Units::c_light),mcPW );
          }
          else
          {
            richHisto1D( HID("ckResTruePion",rad) ) -> fill( deltaTheta );
	    richProfile2D( HID("timeMomentumTruePion",rad) ) -> fill( pTot, (photonTime-trackTime.T()-pathLengthRec/Gaudi::Units::c_light),mcPW );
	    richProfile2D( HID("trackPathMomentumTruePion",rad) ) -> fill( pTot, (trackTime.T()-pathLengthRec/Gaudi::Units::c_light),mcPW );
          }
          
        }
        
      } // scalar loop
      
    }
  }
  
}

//-----------------------------------------------------------------------------

// Declaration of the Algorithm Factory
DECLARE_COMPONENT( SIMDPhotonCherenkovAngles  )

//-----------------------------------------------------------------------------
